# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Rotation(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 24

    # Rotation
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Rotation
    def Yaw(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Rotation
    def Roll(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # Rotation
    def Pitch(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16))

def CreateRotation(builder, yaw, roll, pitch):
    builder.Prep(8, 24)
    builder.PrependFloat64(pitch)
    builder.PrependFloat64(roll)
    builder.PrependFloat64(yaw)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

class MessageBody(object):
    NONE = 0
    ClientHandshake = 1
    ServerHandshake = 2
    Initialize = 3
    State = 4
    Obs = 5
    Command = 6
    Tensor = 7
    Dict = 8
    SensorsDict = 9
    SensorsDataDict = 10
    BoundingBox = 11



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Tensor(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Tensor()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTensor(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def TensorBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # Tensor
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Tensor
    def Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Tensor
    def DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # Tensor
    def DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def DataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Tensor
    def Shape(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Tensor
    def ShapeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Tensor
    def ShapeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def ShapeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def TensorStart(builder): builder.StartObject(2)
def Start(builder):
    return TensorStart(builder)
def TensorAddData(builder, data): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0)
def AddData(builder, data):
    return TensorAddData(builder, data)
def TensorStartDataVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def StartDataVector(builder, numElems):
    return TensorStartDataVector(builder, numElems)
def TensorAddShape(builder, shape): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(shape), 0)
def AddShape(builder, shape):
    return TensorAddShape(builder, shape)
def TensorStartShapeVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartShapeVector(builder, numElems):
    return TensorStartShapeVector(builder, numElems)
def TensorEnd(builder): return builder.EndObject()
def End(builder):
    return TensorEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class VehiclePhysics(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = VehiclePhysics()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsVehiclePhysics(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def VehiclePhysicsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # VehiclePhysics
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # VehiclePhysics
    def TorqueCurve(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VehiclePhysics
    def MaxRpm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def Moi(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def DampingRateFullThrottle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def DampingRateZeroThrottleClutchEngaged(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def DampingRateZeroThrottleClutchDisengaged(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def UseGearAutobox(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # VehiclePhysics
    def GearSwitchTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def ClutchStrength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def FinalRatio(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def ForwardGears(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from communications.isp.ForwardGears import ForwardGears
            obj = ForwardGears()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VehiclePhysics
    def ForwardGearsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # VehiclePhysics
    def ForwardGearsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # VehiclePhysics
    def Mass(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def DragCoefficient(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # VehiclePhysics
    def CenterOfMass(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VehiclePhysics
    def SteeringCurve(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VehiclePhysics
    def SteeringCurveLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # VehiclePhysics
    def SteeringCurveIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        return o == 0

    # VehiclePhysics
    def UseSweepWheelCollision(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # VehiclePhysics
    def Wheels(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from communications.isp.WheelPhysics import WheelPhysics
            obj = WheelPhysics()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VehiclePhysics
    def WheelsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # VehiclePhysics
    def WheelsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        return o == 0

def VehiclePhysicsStart(builder): builder.StartObject(17)
def Start(builder):
    return VehiclePhysicsStart(builder)
def VehiclePhysicsAddTorqueCurve(builder, torqueCurve): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(torqueCurve), 0)
def AddTorqueCurve(builder, torqueCurve):
    return VehiclePhysicsAddTorqueCurve(builder, torqueCurve)
def VehiclePhysicsAddMaxRpm(builder, maxRpm): builder.PrependFloat64Slot(1, maxRpm, 0.0)
def AddMaxRpm(builder, maxRpm):
    return VehiclePhysicsAddMaxRpm(builder, maxRpm)
def VehiclePhysicsAddMoi(builder, moi): builder.PrependFloat64Slot(2, moi, 0.0)
def AddMoi(builder, moi):
    return VehiclePhysicsAddMoi(builder, moi)
def VehiclePhysicsAddDampingRateFullThrottle(builder, dampingRateFullThrottle): builder.PrependFloat64Slot(3, dampingRateFullThrottle, 0.0)
def AddDampingRateFullThrottle(builder, dampingRateFullThrottle):
    return VehiclePhysicsAddDampingRateFullThrottle(builder, dampingRateFullThrottle)
def VehiclePhysicsAddDampingRateZeroThrottleClutchEngaged(builder, dampingRateZeroThrottleClutchEngaged): builder.PrependFloat64Slot(4, dampingRateZeroThrottleClutchEngaged, 0.0)
def AddDampingRateZeroThrottleClutchEngaged(builder, dampingRateZeroThrottleClutchEngaged):
    return VehiclePhysicsAddDampingRateZeroThrottleClutchEngaged(builder, dampingRateZeroThrottleClutchEngaged)
def VehiclePhysicsAddDampingRateZeroThrottleClutchDisengaged(builder, dampingRateZeroThrottleClutchDisengaged): builder.PrependFloat64Slot(5, dampingRateZeroThrottleClutchDisengaged, 0.0)
def AddDampingRateZeroThrottleClutchDisengaged(builder, dampingRateZeroThrottleClutchDisengaged):
    return VehiclePhysicsAddDampingRateZeroThrottleClutchDisengaged(builder, dampingRateZeroThrottleClutchDisengaged)
def VehiclePhysicsAddUseGearAutobox(builder, useGearAutobox): builder.PrependBoolSlot(6, useGearAutobox, 0)
def AddUseGearAutobox(builder, useGearAutobox):
    return VehiclePhysicsAddUseGearAutobox(builder, useGearAutobox)
def VehiclePhysicsAddGearSwitchTime(builder, gearSwitchTime): builder.PrependFloat64Slot(7, gearSwitchTime, 0.0)
def AddGearSwitchTime(builder, gearSwitchTime):
    return VehiclePhysicsAddGearSwitchTime(builder, gearSwitchTime)
def VehiclePhysicsAddClutchStrength(builder, clutchStrength): builder.PrependFloat64Slot(8, clutchStrength, 0.0)
def AddClutchStrength(builder, clutchStrength):
    return VehiclePhysicsAddClutchStrength(builder, clutchStrength)
def VehiclePhysicsAddFinalRatio(builder, finalRatio): builder.PrependFloat64Slot(9, finalRatio, 0.0)
def AddFinalRatio(builder, finalRatio):
    return VehiclePhysicsAddFinalRatio(builder, finalRatio)
def VehiclePhysicsAddForwardGears(builder, forwardGears): builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(forwardGears), 0)
def AddForwardGears(builder, forwardGears):
    return VehiclePhysicsAddForwardGears(builder, forwardGears)
def VehiclePhysicsStartForwardGearsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartForwardGearsVector(builder, numElems):
    return VehiclePhysicsStartForwardGearsVector(builder, numElems)
def VehiclePhysicsAddMass(builder, mass): builder.PrependFloat64Slot(11, mass, 0.0)
def AddMass(builder, mass):
    return VehiclePhysicsAddMass(builder, mass)
def VehiclePhysicsAddDragCoefficient(builder, dragCoefficient): builder.PrependFloat64Slot(12, dragCoefficient, 0.0)
def AddDragCoefficient(builder, dragCoefficient):
    return VehiclePhysicsAddDragCoefficient(builder, dragCoefficient)
def VehiclePhysicsAddCenterOfMass(builder, centerOfMass): builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(centerOfMass), 0)
def AddCenterOfMass(builder, centerOfMass):
    return VehiclePhysicsAddCenterOfMass(builder, centerOfMass)
def VehiclePhysicsAddSteeringCurve(builder, steeringCurve): builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(steeringCurve), 0)
def AddSteeringCurve(builder, steeringCurve):
    return VehiclePhysicsAddSteeringCurve(builder, steeringCurve)
def VehiclePhysicsStartSteeringCurveVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartSteeringCurveVector(builder, numElems):
    return VehiclePhysicsStartSteeringCurveVector(builder, numElems)
def VehiclePhysicsAddUseSweepWheelCollision(builder, useSweepWheelCollision): builder.PrependBoolSlot(15, useSweepWheelCollision, 0)
def AddUseSweepWheelCollision(builder, useSweepWheelCollision):
    return VehiclePhysicsAddUseSweepWheelCollision(builder, useSweepWheelCollision)
def VehiclePhysicsAddWheels(builder, wheels): builder.PrependUOffsetTRelativeSlot(16, flatbuffers.number_types.UOffsetTFlags.py_type(wheels), 0)
def AddWheels(builder, wheels):
    return VehiclePhysicsAddWheels(builder, wheels)
def VehiclePhysicsStartWheelsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartWheelsVector(builder, numElems):
    return VehiclePhysicsStartWheelsVector(builder, numElems)
def VehiclePhysicsEnd(builder): return builder.EndObject()
def End(builder):
    return VehiclePhysicsEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Obs(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Obs()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsObs(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ObsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # Obs
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Obs
    def EgoLocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Location import Location
            obj = Location()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Obs
    def EgoRotation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Rotation import Rotation
            obj = Rotation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Obs
    def Speed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # Obs
    def FrontImage(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Obs
    def BirdviewImage(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Obs
    def Command(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Obs
    def CompactVector(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Obs
    def PrevAction(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Action import Action
            obj = Action()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Obs
    def SensorData(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.SensorsDataDict import SensorsDataDict
            obj = SensorsDataDict()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def ObsStart(builder): builder.StartObject(9)
def Start(builder):
    return ObsStart(builder)
def ObsAddEgoLocation(builder, egoLocation): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(egoLocation), 0)
def AddEgoLocation(builder, egoLocation):
    return ObsAddEgoLocation(builder, egoLocation)
def ObsAddEgoRotation(builder, egoRotation): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(egoRotation), 0)
def AddEgoRotation(builder, egoRotation):
    return ObsAddEgoRotation(builder, egoRotation)
def ObsAddSpeed(builder, speed): builder.PrependFloat64Slot(2, speed, 0.0)
def AddSpeed(builder, speed):
    return ObsAddSpeed(builder, speed)
def ObsAddFrontImage(builder, frontImage): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(frontImage), 0)
def AddFrontImage(builder, frontImage):
    return ObsAddFrontImage(builder, frontImage)
def ObsAddBirdviewImage(builder, birdviewImage): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(birdviewImage), 0)
def AddBirdviewImage(builder, birdviewImage):
    return ObsAddBirdviewImage(builder, birdviewImage)
def ObsAddCommand(builder, command): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(command), 0)
def AddCommand(builder, command):
    return ObsAddCommand(builder, command)
def ObsAddCompactVector(builder, compactVector): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(compactVector), 0)
def AddCompactVector(builder, compactVector):
    return ObsAddCompactVector(builder, compactVector)
def ObsAddPrevAction(builder, prevAction): builder.PrependStructSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(prevAction), 0)
def AddPrevAction(builder, prevAction):
    return ObsAddPrevAction(builder, prevAction)
def ObsAddSensorData(builder, sensorData): builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(sensorData), 0)
def AddSensorData(builder, sensorData):
    return ObsAddSensorData(builder, sensorData)
def ObsEnd(builder): return builder.EndObject()
def End(builder):
    return ObsEnd(builder)


"""
To improve performance assertions could be commented
uncomment to test and comment for deployment
TODO:
    Write Docstring
    Add logger for exceptions
    Add type hints
"""

from typing import Dict

import cv2
import numpy as np
import torch

from iai_common.communications.isp import Action as isp_Action
from iai_common.communications.isp import BoundingBox as isp_BoundingBox
from iai_common.communications.isp import ClientHandshake as isp_ClientHandshake
from iai_common.communications.isp import Command as isp_Command
from iai_common.communications.isp import Dict as isp_Dict
from iai_common.communications.isp import Info as isp_Info
from iai_common.communications.isp import Initialize as isp_Initialize
from iai_common.communications.isp import Location as isp_Location
from iai_common.communications.isp import Message as isp_Message
from iai_common.communications.isp import MessageBody as isp_MessageBody
from iai_common.communications.isp import Obs as isp_Obs
from iai_common.communications.isp import Resolution as isp_Resolution
from iai_common.communications.isp import Rotation as isp_Rotation
from iai_common.communications.isp import Sensor as isp_Sensor
from iai_common.communications.isp import SensorData as isp_SensorData
from iai_common.communications.isp import SensorsDataDict as isp_SensorsDataDict
from iai_common.communications.isp import SensorsDict as isp_SensorsDict
from iai_common.communications.isp import ServerHandshake as isp_ServerHandshake
from iai_common.communications.isp import State as isp_State
from iai_common.communications.isp import Step as isp_Step
from iai_common.communications.isp import Tensor as isp_Tensor
from iai_common.communications.isp import WorldParameters as isp_WorldParameters
from iai_common.communications.isp import ScenarioParameters as isp_ScenarioParameters
import collections


class SensorSettings:
    Available_Sensors = ['camera']
    Available_Camera_types = ['rgb-camera', 'segmentation', 'depth-camera']
    Location = collections.namedtuple('Location', ['x', 'y', 'z'])
    Rotation = collections.namedtuple('Rotation', ['yaw', 'pitch', 'roll'])
    Resolution = collections.namedtuple('Resolution', ['width', 'height'])
    Available_Tracked_Actors = ['vehicles', 'pedestrians', 'traffic-lights', 'parked_vehicles']
    Available_Reference_Frame = ['carla']  # 'opendrive', 'geolocation' will be added
    Default_Settings = {
        'camera': {
            'sensor_type': 'camera',
            'camera_type': 'rgb-camera',
            'bounding_box': True,
            'track_actor_types': Available_Tracked_Actors,  # or 'all'
            'show_bounding_boxes': True,
            'world_sensor': False,
            'location': Location(x=1.6, z=1.7, y=0),
            'rotation': Rotation(yaw=0, pitch=0, roll=0),
            'resolution': Resolution(200, 88),
            'fov': 90.0,
        },
        'boundingbox': {
            'sensor_type': 'boundingbox',
            'track_actor_types': Available_Tracked_Actors,  # or 'all'
            'world_sensor': False,
            'location': Location(x=0, z=0, y=0),
            'rotation': Rotation(yaw=0, pitch=0, roll=0),
            'frame_of_reference': 'carla',
            # if world_sensor=True returns coordinated based on either 'carla', 'opendrive', 'geolocation'
            'attach_to_actor': 'ego',
            'radius': 100000,  # If not specified all world is considered
            'occlusion': False,
        },
    }


def get_message_body(message_buffer):
    message = isp_Message.Message.GetRootAsMessage(message_buffer, 0)
    body_type = message.BodyType()
    if body_type == isp_MessageBody.MessageBody().ServerHandshake:
        message_body = isp_ServerHandshake.ServerHandshake()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        system_name = message_body.SystemName().decode("utf-8")
        model_name = message_body.ModelName().decode("utf-8")
        return "ServerHandshake", (system_name, model_name)
    elif body_type == isp_MessageBody.MessageBody().ClientHandshake:
        message_body = isp_ClientHandshake.ClientHandshake()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        return "ClientHandshake", (message_body.SystemName().decode("utf-8"))
    elif body_type == isp_MessageBody.MessageBody().Tensor:
        message_body = isp_Tensor.Tensor()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        data = message_body.DataAsNumpy()
        shape = message_body.ShapeAsNumpy()
        return "Tensor", (data.reshape(shape))
    elif body_type == isp_MessageBody.MessageBody().Command:
        message_body = isp_Command.Command()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        return "Command", unpack_command(message_body)
    elif body_type == isp_MessageBody.MessageBody().Initialize:
        message_body = isp_Initialize.Initialize()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        return "Initialize", unpack_initialize(message_body)
    elif body_type == isp_MessageBody.MessageBody().Dict:
        message_body = isp_Dict.Dict()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        return "Dict", unpack_dict(message_body)
    elif body_type == isp_MessageBody.MessageBody().Obs:
        message_body = isp_Obs.Obs()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        return "Obs", unpack_obs(message_body)
    elif body_type == isp_MessageBody.MessageBody().State:
        message_body = isp_State.State()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        return "State", unpack_state(message_body)
    elif body_type == isp_MessageBody.MessageBody().SensorsDict:
        message_body = isp_SensorsDict.SensorsDict()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        return "Sensors", unpack_sensors_dict(message_body)
    elif body_type == isp_MessageBody.MessageBody().SensorsDataDict:
        message_body = isp_SensorsDataDict.SensorsDataDict()
        message_body.Init(message.Body().Bytes, message.Body().Pos)
        return "Sensors_Data", unpack_sensor_data_dict(message_body)
    else:
        raise RuntimeError(
            "isp (Python):\
            Received unexpected message body type: {}".format(body_type)
        )


def build_message(builder, message_body, body_type):
    isp_Message.MessageStart(builder)
    isp_Message.MessageAddBodyType(builder,
                                   body_type)
    isp_Message.MessageAddBody(builder, message_body)
    message = isp_Message.MessageEnd(builder)
    builder.Finish(message)
    return builder.Output()


def build_clienthandshake(builder, system_name):
    assert type(system_name) == str
    system_name = builder.CreateString(system_name)
    isp_ClientHandshake.ClientHandshakeStart(builder)
    isp_ClientHandshake.ClientHandshakeAddSystemName(builder, system_name)

    # return = isp_ClientHandshake.ClientHandshakeEnd(builder)
    message_body = isp_ClientHandshake.ClientHandshakeEnd(builder)
    return build_message(builder, message_body,
                         isp_MessageBody.MessageBody().ClientHandshake)


def build_serverhandshake(builder, system_name, model_name):
    assert type(system_name) == str
    assert type(model_name) == str
    system_name = builder.CreateString(system_name)
    model_name = builder.CreateString(model_name)
    isp_ServerHandshake.ServerHandshakeStart(builder)
    isp_ServerHandshake.ServerHandshakeAddSystemName(builder, system_name)
    isp_ServerHandshake.ServerHandshakeAddModelName(builder, model_name)
    message_body = isp_ServerHandshake.ServerHandshakeEnd(builder)
    return build_message(builder, message_body,
                         isp_MessageBody.MessageBody().ServerHandshake)


def build_tensor(builder, data: np.ndarray, get_as_message=False):
    assert ((type(data) == np.ndarray) or
            (type(data) == np.matrix))
    shape = list(data.shape)
    data = np.asarray(data).flatten().tolist()
    # pack data
    isp_Tensor.TensorStartDataVector(builder, len(data))
    for d in reversed(data):
        builder.PrependFloat64(d)
    data = builder.EndVector()
    # pack shape
    isp_Tensor.TensorStartShapeVector(builder, len(shape))
    for s in reversed(shape):
        builder.PrependInt32(s)
    shape = builder.EndVector()
    isp_Tensor.TensorStart(builder)
    isp_Tensor.TensorAddData(builder, data)
    isp_Tensor.TensorAddShape(builder, shape)
    if get_as_message:
        message_body = isp_Tensor.TensorEnd(builder)
        return build_message(builder, message_body,
                             isp_MessageBody.MessageBody().Tensor)
    else:
        return isp_Tensor.TensorEnd(builder)


def unpack_tensor(message_body):
    data = message_body.DataAsNumpy()
    shape = message_body.ShapeAsNumpy()
    return data.reshape(shape)


def build_dict(builder, in_dict: Dict, get_as_message=False):
    encoded_keys = []
    encoded_values = []
    for key in in_dict.keys():
        encoded_keys.append(builder.CreateString(key))
    for value in in_dict.values():
        encoded_values.append(build_tensor(builder, value))

    isp_Dict.DictStartKeysVector(builder, len(in_dict))
    for key in reversed(encoded_keys):
        builder.PrependUOffsetTRelative(key)
    serialized_keys = builder.EndVector()
    isp_Dict.DictStartValuesVector(builder, len(in_dict))
    for value in reversed(encoded_values):
        builder.PrependUOffsetTRelative(value)
    serialized_values = builder.EndVector()

    isp_Dict.DictStart(builder)
    isp_Dict.DictAddKeys(builder, serialized_keys)
    isp_Dict.DictAddValues(builder, serialized_values)
    if get_as_message:
        message_body = isp_Dict.DictEnd(builder)
        return build_message(builder, message_body,
                             isp_MessageBody.MessageBody().Dict)
    else:
        return isp_Dict.DictEnd(builder)


def unpack_dict(message_body):
    decoded_keys = []
    decoded_values = []
    for l in range(message_body.KeysLength()):
        decoded_keys.append(message_body.Keys(l).decode())
        decoded_values.append(unpack_tensor(message_body.Values(l)))
    decoded_dict = {decoded_keys[i]: decoded_values[i]
                    for i in range(message_body.KeysLength())}
    return decoded_dict


def build_command(builder, cmd, data=None):
    if data is None:
        data = {}
    command_text = builder.CreateString(cmd)
    if 'tensor' in data.keys():
        packed_tensor = build_tensor(builder, data['tensor'])
    if 'step' in data.keys():
        packed_step = build_step(builder, data['step'])
    isp_Command.CommandStart(builder)
    isp_Command.CommandAddCmd(builder, command_text)
    if 'tensor' in data.keys():
        isp_Command.CommandAddTensorArgs(builder, packed_tensor)
    if 'step' in data.keys():
        isp_Command.CommandAddStep(builder, packed_step)
    message_body = isp_Command.CommandEnd(builder)
    return build_message(builder, message_body,
                         isp_MessageBody.MessageBody().Command)


def unpack_command(message_body):
    cmd = message_body.Cmd().decode("utf-8")
    body = {}
    if message_body.TensorArgs() is not None:
        data = message_body.TensorArgs().DataAsNumpy()
        shape = message_body.TensorArgs().ShapeAsNumpy()
        body['tensor'] = data.reshape(shape)
    if message_body.Step() is not None:
        body['step'] = unpack_step(message_body.Step())
    return cmd, body


def build_initialize(builder, scenario, scenario_parameters=None,
                     world_parameters=None, vehicle_physics=None,
                     sensors=None):
    scenario_name = builder.CreateString(scenario)
    if scenario_parameters is not None:
        isp_ScenarioParameters.ScenarioParametersStart(builder)
        if scenario_parameters['camera_location_variation'] is not None:
            isp_ScenarioParameters.ScenarioParametersAddCameraLocationVariation(builder, isp_Location.CreateLocation(
                builder, scenario_parameters['camera_location_variation'].x,
                scenario_parameters['camera_location_variation'].y, scenario_parameters['camera_location_variation'].z))
        if scenario_parameters['camera_rotation_variation'] is not None:
            isp_ScenarioParameters.ScenarioParametersAddCameraOrientationVariation(builder, isp_Rotation.CreateRotation(
                builder, scenario_parameters['camera_rotation_variation'].yaw,
                scenario_parameters['camera_rotation_variation'].roll,
                scenario_parameters['camera_rotation_variation'].pitch))
        scenarioparam = isp_ScenarioParameters.ScenarioParametersEnd(builder)
    if vehicle_physics is not None:
        pass
    if world_parameters is not None:
        if world_parameters['carlatown'] is not None:
            carlatown = builder.CreateString(world_parameters['carlatown'])
        isp_WorldParameters.WorldParametersStart(builder)
        if world_parameters['carlatown'] is not None:
            isp_WorldParameters.WorldParametersAddCarlatown(builder, carlatown)
            worldparam = isp_WorldParameters.WorldParametersEnd(builder)
    if sensors is not None:
        serialized_sensors = build_sensor_dict(builder, sensors, get_as_message=False)
    isp_Initialize.InitializeStart(builder)
    isp_Initialize.InitializeAddScenario(builder, scenario_name)
    if scenario_parameters is not None:
        isp_Initialize.InitializeAddScenarioParameters(builder, scenarioparam)
    if vehicle_physics is not None:
        pass
    if world_parameters is not None:
        isp_Initialize.InitializeAddWorldParameters(builder, worldparam)
    if sensors is not None:
        isp_Initialize.InitializeAddSensors(builder, serialized_sensors)
    message_body = isp_Initialize.InitializeEnd(builder)
    return build_message(builder, message_body,
                         isp_MessageBody.MessageBody().Initialize)


def unpack_initialize(message_body):
    scenario = message_body.Scenario().decode("utf-8")
    world_parameters = None
    sensors = None
    vehicle_physics = None
    scenario_parameters = None
    if message_body.WorldParameters() is not None:
        world_parameters = {}
        if message_body.WorldParameters().Carlatown() is not None:
            world_parameters['carlatown'] = message_body.WorldParameters().Carlatown().decode("utf-8")
    if message_body.ScenarioParameters() is not None:
        scenario_parameters = {}
        if message_body.ScenarioParameters().CameraLocationVariation() is not None:
            scenario_parameters['camera_location_variation'] = SensorSettings.Location(
                x=message_body.ScenarioParameters().CameraLocationVariation().X(),
                z=message_body.ScenarioParameters().CameraLocationVariation().Z(),
                y=message_body.ScenarioParameters().CameraLocationVariation().Y())
        if message_body.ScenarioParameters().CameraOrientationVariation() is not None:
            scenario_parameters['camera_rotation_variation'] = SensorSettings.Rotation(
                yaw=message_body.ScenarioParameters().CameraOrientationVariation().Yaw(),
                pitch=message_body.ScenarioParameters().CameraOrientationVariation().Pitch(),
                roll=message_body.ScenarioParameters().CameraOrientationVariation().Roll())
    if message_body.Sensors() is not None:
        sensors = unpack_sensors_dict(message_body.Sensors())
    body = {'scenario': scenario, 'world_parameters': world_parameters,
            'vehicle_physics': vehicle_physics,
            'scenario_parameters': scenario_parameters, 'sensors': sensors}
    return body


def build_obs(builder, obs: Dict, get_as_message: bool = False, compress=True):
    """
    speed: float front_image: torch.Tensor
    birdview_image: torch.Tensor, command: str,
    compact_vector: np.ndarray, prev_action: tuple,
    """
    command = builder.CreateString(obs['command'])

    if compress:
        im = (obs['front_image'].squeeze().permute(1, 2, 0).numpy() * 255).astype(np.uint8)
        img_encode = cv2.imencode('.jpg', im)[1].reshape([-1, 1])
        front_image = build_tensor(builder, img_encode)
        im = (obs['birdview_image'].squeeze().permute(1, 2, 0).numpy() * 255).astype(np.uint8)
        img_encode = cv2.imencode('.jpg', im)[1].reshape([-1, 1])
        birdview_image = build_tensor(builder, img_encode)
    else:
        front_image = build_tensor(builder, obs['front_image'].numpy())
        birdview_image = build_tensor(builder, obs['birdview_image'].numpy())
    compact_vector = build_tensor(builder, obs['compact_vector'])
    if 'sensor_data' in obs.keys():
        sensor_data = build_sensor_data_dict(builder, obs['sensor_data'])
    # prev_action = build_tensor(builder, np.array(obs['prev_action']))
    isp_Obs.ObsStart(builder)
    isp_Obs.ObsAddSpeed(builder, obs['speed'])
    isp_Obs.ObsAddFrontImage(builder, front_image)
    isp_Obs.ObsAddBirdviewImage(builder, birdview_image)
    isp_Obs.ObsAddCommand(builder, command)
    isp_Obs.ObsAddCompactVector(builder, compact_vector)
    isp_Obs.ObsAddPrevAction(builder, isp_Action.CreateAction(
        builder, obs['prev_action'][0], obs['prev_action'][1]))
    if 'sensor_data' in obs.keys():
        isp_Obs.ObsAddSensorData(builder, sensor_data)
    message_body = isp_Obs.ObsEnd(builder)
    if get_as_message:
        return build_message(builder, message_body,
                             isp_MessageBody.MessageBody().Obs)
    else:
        return message_body


def unpack_obs(message_body, compress=True):
    obs = {}
    obs['speed'] = message_body.Speed()
    if compress:
        rec_im = unpack_tensor(message_body.FrontImage()).astype(np.uint8).squeeze()
        obs['front_image'] = cv2.imdecode(rec_im, cv2.IMREAD_COLOR)
        rec_im = unpack_tensor(message_body.BirdviewImage()).astype(np.uint8).squeeze()
        obs['birdview_image'] = cv2.imdecode(rec_im, cv2.IMREAD_COLOR)
    else:
        obs['front_image'] = torch.Tensor(unpack_tensor(message_body.FrontImage()).copy())
        obs['birdview_image'] = torch.Tensor(unpack_tensor(message_body.BirdviewImage()).copy())
    obs['command'] = message_body.Command().decode()
    obs['compact_vector'] = unpack_tensor(message_body.CompactVector())
    obs['prev_action'] = (message_body.PrevAction().Acceleration(),
                          message_body.PrevAction().Steering())
    if message_body.SensorData() is not None:
        obs['sensor_data'] = unpack_sensor_data_dict(message_body.SensorData())
    return obs


def build_info(builder, info: Dict):
    """
      invasion: bool;
      collision: bool;
      gear: int;
      expert_action: isp_Action;
      outcome: Tensor;                  // TODO: Not sure what it should be
      next_waypoint_distance: float;
      distance_to_goal: float;
      waypoints_found: int;
      waypoints_total: int;
      waypoints_remaining: int;
   """
    outcome = build_tensor(builder, info['outcome'])

    isp_Info.InfoStart(builder)
    isp_Info.InfoAddInvasion(builder, info['invasion'])
    isp_Info.InfoAddCollision(builder, info['collision'])
    isp_Info.InfoAddGear(builder, info['gear'])
    isp_Info.InfoAddExpertAction(builder, isp_Action.CreateAction(
        builder, info['expert_action'][0], info['expert_action'][1]))
    isp_Info.InfoAddOutcome(builder, outcome)
    if 'next_waypoint_distance' in info.keys():
        isp_Info.InfoAddNextWaypointDistance(builder, info['next_waypoint_distance'])
    if 'distance_to_goal' in info.keys():
        isp_Info.InfoAddDistanceToGoal(builder, info['distance_to_goal'])
    if 'waypoints_found' in info.keys():
        isp_Info.InfoAddWaypointsFound(builder, info['waypoints_found'])
    if 'waypoints_total' in info.keys():
        isp_Info.InfoAddWaypointsTotal(builder, info['waypoints_total'])
    if 'waypoints_remaining' in info.keys():
        isp_Info.InfoAddWaypointsRemaining(builder, info['waypoints_remaining'])
    return isp_Info.InfoEnd(builder)


def unpack_info(message_body):
    info = {'invasion': message_body.Invasion(), 'collision': message_body.Collision(), 'gear': message_body.Gear(),
            'expert_action': (message_body.ExpertAction().Acceleration(),
                              message_body.ExpertAction().Steering()), 'outcome': unpack_tensor(message_body.Outcome()),
            'next_waypoint_distance': message_body.NextWaypointDistance(),
            'distance_to_goal': message_body.DistanceToGoal(), 'waypoints_found': message_body.WaypointsFound(),
            'waypoints_total': message_body.WaypointsTotal(), 'waypoints_remaining': message_body.WaypointsRemaining()}
    return info


def build_state(builder, obs, reward, done, info):
    packed_obs = build_obs(builder, obs)
    packed_info = build_info(builder, info)
    isp_State.StateStart(builder)
    isp_State.StateAddObs(builder, packed_obs)
    isp_State.StateAddReward(builder, reward)
    isp_State.StateAddDone(builder, done)
    isp_State.StateAddInfo(builder, packed_info)
    message_body = isp_State.StateEnd(builder)
    return build_message(builder, message_body,
                         isp_MessageBody.MessageBody().State)


def unpack_state(message_body):
    obs = unpack_obs(message_body.Obs())
    reward = message_body.Reward()
    done = message_body.Done()
    info = unpack_info(message_body.Info())
    return {'obs': obs, 'reward': reward, 'done': done, 'info': info}


def build_step(builder, action):
    isp_Step.StepStart(builder)
    isp_Step.StepAddAction(builder, isp_Action.CreateAction(builder, action[0],
                                                            action[1]))
    return isp_Step.StepEnd(builder)


def unpack_step(message_body):
    action = (message_body.Action().Acceleration(),
              message_body.Action().Steering())
    return action


def build_sensor(builder, sensor_name: str, sensor: Dict):
    name = builder.CreateString(sensor_name)
    sensor_type = builder.CreateString(sensor['sensor_type'])
    if 'camera_type' in sensor.keys():
        camera_type = builder.CreateString(sensor['camera_type'])
    if 'track_actor_types' in sensor.keys():
        track_actor_types = []
        for actor in sensor['track_actor_types']:
            track_actor_types.append(builder.CreateString(actor))
        isp_Sensor.SensorStartTrackActorTypesVector(builder, len(track_actor_types))
        for actor in reversed(track_actor_types):
            builder.PrependUOffsetTRelative(actor)
        serialized_actor = builder.EndVector()
    if 'attach_to_actor' in sensor.keys():
        attach_to_actor = builder.CreateString(sensor['attach_to_actor'])
    if 'frame_of_reference' in sensor.keys():
        frame_of_reference = builder.CreateString(sensor['frame_of_reference'])

    isp_Sensor.SensorStart(builder)
    isp_Sensor.SensorAddSensorName(builder, name)
    isp_Sensor.SensorAddSensorTypes(builder, sensor_type)
    if 'camera_type' in sensor.keys():
        isp_Sensor.SensorAddCameraType(builder, camera_type)
    if 'bounding_box' in sensor.keys():
        isp_Sensor.SensorAddBoundingBox(builder, sensor['bounding_box'])
    if 'track_actor_types' in sensor.keys():
        isp_Sensor.SensorAddTrackActorTypes(builder, serialized_actor)
    if 'show_bounding_boxes' in sensor.keys():
        isp_Sensor.SensorAddShowBoundingBoxes(builder, sensor['show_bounding_boxes'])
    if 'world_sensor' in sensor.keys():
        isp_Sensor.SensorAddWorldSensor(builder, sensor['world_sensor'])
    if 'location' in sensor.keys():
        isp_Sensor.SensorAddSensorLocation(builder, isp_Location.CreateLocation(
            builder, sensor['location'].x, sensor['location'].y, sensor['location'].z))
    if 'rotation' in sensor.keys():
        isp_Sensor.SensorAddSensorRotation(builder, isp_Rotation.CreateRotation(
            builder, sensor['rotation'].yaw, sensor['rotation'].roll, sensor['rotation'].pitch))
    if 'resolution' in sensor.keys():
        isp_Sensor.SensorAddSensorResolution(builder, isp_Resolution.CreateResolution(
            builder, sensor['resolution'].width, sensor['resolution'].height))
    if 'radius' in sensor.keys():
        isp_Sensor.SensorAddRadius(builder, sensor['radius'])
    if 'fov' in sensor.keys():
        isp_Sensor.SensorAddFov(builder, sensor['fov'])
    if 'attach_to_actor' in sensor.keys():
        isp_Sensor.SensorAddAttachToActor(builder, attach_to_actor)
    if 'frame_of_reference' in sensor.keys():
        isp_Sensor.SensorAddFrameOfReference(builder, frame_of_reference)
    if 'occlusion' in sensor.keys():
        isp_Sensor.SensorAddOcclusion(builder, sensor['occlusion'])

    return isp_Sensor.SensorEnd(builder)


def unpack_sensor(message_body):
    sensor = {}
    name = message_body.SensorName().decode("utf-8")
    sensor['sensor_type'] = message_body.SensorTypes().decode("utf-8")
    if message_body.CameraType() is not None:
        sensor['camera_type'] = message_body.CameraType().decode("utf-8")
    if message_body.BoundingBox() is not None:
        sensor['bounding_box'] = message_body.BoundingBox()
    if not message_body.TrackActorTypesIsNone():
        track_actor_types = []
        for l in range(message_body.TrackActorTypesLength()):
            track_actor_types.append(message_body.TrackActorTypes(l).decode())
        sensor['track_actor_types'] = track_actor_types
    if message_body.ShowBoundingBoxes() is not None:
        sensor['show_bounding_boxes'] = message_body.ShowBoundingBoxes()
    if message_body.WorldSensor() is not None:
        sensor['world_sensor'] = message_body.WorldSensor()
    if message_body.SensorResolution() is not None:
        sensor['resolution'] = SensorSettings.Resolution(
            width=message_body.SensorResolution().Width(),
            height=message_body.SensorResolution().Height())
    if message_body.SensorLocation() is not None:
        sensor['location'] = SensorSettings.Location(
            x=message_body.SensorLocation().X(),
            z=message_body.SensorLocation().Z(),
            y=message_body.SensorLocation().Y())
    if message_body.SensorRotation() is not None:
        sensor['rotation'] = SensorSettings.Rotation(
            yaw=message_body.SensorRotation().Yaw(),
            pitch=message_body.SensorRotation().Pitch(),
            roll=message_body.SensorRotation().Roll())
    if message_body.Fov() != 0.0:
        sensor['fov'] = message_body.Fov()
    if message_body.Radius() != 0.0:
        sensor['radius'] = message_body.Radius()
    if message_body.FrameOfReference() is not None:
        sensor['frame_of_reference'] = message_body.FrameOfReference().decode("utf-8")
    if message_body.AttachToActor() is not None:
        sensor['attach_to_actor'] = message_body.AttachToActor().decode("utf-8")
    if message_body.Occlusion() is not None:
        sensor['occlusion'] = message_body.Occlusion()
    return name, sensor


def build_sensor_dict(builder, sensors_dict: Dict, get_as_message=False):
    sensor_array = []
    for sensor in sensors_dict:
        sensor_array.append(build_sensor(
            builder, sensor, sensors_dict[sensor]))
    isp_SensorsDict.SensorsDictStartSensorsVector(builder, len(sensor_array))
    for sensor in reversed(sensor_array):
        builder.PrependUOffsetTRelative(sensor)
    serialized_sensor = builder.EndVector()
    isp_SensorsDict.SensorsDictStart(builder)
    isp_SensorsDict.SensorsDictAddSensors(builder, serialized_sensor)
    if get_as_message:
        message_body = isp_SensorsDict.SensorsDictEnd(builder)
        return build_message(builder, message_body,
                             isp_MessageBody.MessageBody().SensorsDict)
    else:
        return isp_SensorsDict.SensorsDictEnd(builder)


def unpack_sensors_dict(message_body):
    sensors_dict = {}
    for l in range(message_body.SensorsLength()):
        name, sns = unpack_sensor(message_body.Sensors(l))
        sensors_dict[name] = sns
    return sensors_dict


def build_sensor_data(builder, sensor_name: str, data: Dict, get_as_message=False, compress=True):
    name = builder.CreateString(sensor_name)
    no_id_bbox = None
    bbox = None
    image = None
    for key in data:
        if key == 'image':
            if compress:
                img_encode = cv2.imencode('.jpg', data[key])[1].reshape([-1, 1])
                image = build_tensor(builder, img_encode)
            else:
                image = build_tensor(builder, data[key])
        elif key == 'bounding_boxes':
            bbox = []
            for actor in data[key]:
                bbox.append(build_bounding_box(builder, actor, data[key][actor]))
            pass
        elif key == 'no_id_bounding_boxes':
            for actor in data[key].keys():
                data[key][actor] = np.array(data[key][actor])
            no_id_bbox = build_dict(builder, data[key])
        else:
            raise RuntimeError(
                "isp (Python):\
                unexpected sensor data type: {}".format(key))

    if bbox is not None:
        isp_SensorData.SensorDataStartBoundingBoxVector(builder, len(bbox))
        for box in bbox:
            builder.PrependUOffsetTRelative(box)
        serialized_bbox = builder.EndVector()

    isp_SensorData.SensorDataStart(builder)
    isp_SensorData.SensorDataAddSensorName(builder, name)
    if image is not None:
        isp_SensorData.SensorDataAddImage(builder, image)
    if bbox is not None:
        isp_SensorData.SensorDataAddBoundingBox(builder, serialized_bbox)
    return isp_SensorData.SensorDataEnd(builder)


def unpack_sensor_data(message_body, compress=True):
    sensor_data = {}
    name = message_body.SensorName().decode("utf-8")
    if message_body.Image() is not None:
        if compress:
            rec_im = unpack_tensor(message_body.Image()).astype(np.uint8).squeeze()
            im = cv2.imdecode(rec_im, cv2.IMREAD_COLOR)
            sensor_data['image'] = im
        else:
            sensor_data['image'] = unpack_tensor(message_body.Image())
    if not message_body.BoundingBoxIsNone():
        sensor_data['bounding_boxes'] = {}
        for l in range(message_body.BoundingBoxLength()):
            actor_type, data_list = unpack_bounding_box(message_body.BoundingBox(l))
            sensor_data['bounding_boxes'][actor_type] = data_list
    return name, sensor_data


def build_sensor_data_dict(builder, sensors_data_dict: Dict, get_as_message=False):
    sensor_data_array = []
    for sensor in sensors_data_dict:
        sensor_data_array.append(build_sensor_data(
            builder, sensor, sensors_data_dict[sensor]))
    isp_SensorsDataDict.SensorsDataDictStartSensorsDataVector(builder, len(sensor_data_array))
    for sensor in reversed(sensor_data_array):
        builder.PrependUOffsetTRelative(sensor)
    serialized_sensor_data = builder.EndVector()
    isp_SensorsDataDict.SensorsDataDictStart(builder)
    isp_SensorsDataDict.SensorsDataDictAddSensorsData(builder, serialized_sensor_data)
    message_body = isp_SensorsDataDict.SensorsDataDictEnd(builder)
    if get_as_message:
        return build_message(builder, message_body,
                             isp_MessageBody.MessageBody().SensorsDataDict)
    else:
        return message_body


def unpack_sensor_data_dict(message_body):
    sensors_data_dict = {}
    for i in range(message_body.SensorsDataLength()):
        name, data = unpack_sensor_data(message_body.SensorsData(i))
        sensors_data_dict[name] = data
    return sensors_data_dict


def build_bounding_box(builder, actor_type: str, data_list: Dict):
    name = builder.CreateString(actor_type)
    encoded_ids = []
    encoded_cords = []
    encoded_states = []
    encoded_occlusion = []
    for item_id, item in data_list.items():
        encoded_cords.append(build_tensor(builder, item['cords']))
        encoded_ids.append(int(item_id))
        if 'state' in item.keys():
            encoded_states.append(builder.CreateString(item['state']))
        if 'occlusion' in item.keys():
            encoded_occlusion.append(item['occlusion'])

    isp_BoundingBox.BoundingBoxStartIdsVector(builder, len(encoded_ids))
    for int_id in reversed(encoded_ids):
        builder.PrependUint64(int_id)
    serialized_ids = builder.EndVector()
    isp_BoundingBox.BoundingBoxStartCoordinatesVector(builder, len(encoded_cords))
    for cord in reversed(encoded_cords):
        builder.PrependUOffsetTRelative(cord)
    serialized_cords = builder.EndVector()
    if len(encoded_states) > 0:
        isp_BoundingBox.BoundingBoxStartStateVector(builder, len(encoded_states))
        for cord in reversed(encoded_states):
            builder.PrependUOffsetTRelative(cord)
        serialized_states = builder.EndVector()
    if len(encoded_occlusion) > 0:
        isp_BoundingBox.BoundingBoxStartOcclusionVector(builder, len(encoded_occlusion))
        for ocl in reversed(encoded_occlusion):
            builder.PrependFloat64(ocl)
        serialized_occlusion = builder.EndVector()

    isp_BoundingBox.BoundingBoxStart(builder)
    isp_BoundingBox.BoundingBoxAddActorType(builder, name)
    isp_BoundingBox.BoundingBoxAddIds(builder, serialized_ids)
    isp_BoundingBox.BoundingBoxAddCoordinates(builder, serialized_cords)
    if len(encoded_states) > 0:
        isp_BoundingBox.BoundingBoxAddState(builder, serialized_states)
    if len(encoded_occlusion) > 0:
        isp_BoundingBox.BoundingBoxAddOcclusion(builder, serialized_occlusion)

    return isp_BoundingBox.BoundingBoxEnd(builder)


def unpack_bounding_box(message_body):
    decoded = {}
    actor_type = message_body.ActorType().decode("utf-8")
    for l in range(message_body.IdsLength()):
        decoded_id = message_body.Ids(l)
        decoded_cords = unpack_tensor(message_body.Coordinates(l))
        decoded_state = None
        decoded_occlusion = None
        if not message_body.StateIsNone():
            decoded_state = message_body.State(l).decode("utf-8")
        if not message_body.OcclusionIsNone():
            decoded_occlusion = message_body.Occlusion(l)
        decoded[decoded_id] = {}
        decoded[decoded_id] = {'cords': decoded_cords,
                               'state': decoded_state,
                               'occlusion': decoded_occlusion}
    return actor_type, decoded



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Sensor(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Sensor()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSensor(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def SensorBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # Sensor
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Sensor
    def SensorName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Sensor
    def SensorTypes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Sensor
    def CameraType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Sensor
    def BoundingBox(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Sensor
    def TrackActorTypes(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Sensor
    def TrackActorTypesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Sensor
    def TrackActorTypesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Sensor
    def ShowBoundingBoxes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Sensor
    def WorldSensor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Sensor
    def SensorLocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Location import Location
            obj = Location()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Sensor
    def SensorRotation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Rotation import Rotation
            obj = Rotation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Sensor
    def SensorResolution(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Resolution import Resolution
            obj = Resolution()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Sensor
    def FrameOfReference(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Sensor
    def Fov(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # Sensor
    def AttachToActor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Sensor
    def Radius(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # Sensor
    def Occlusion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def SensorStart(builder): builder.StartObject(15)
def Start(builder):
    return SensorStart(builder)
def SensorAddSensorName(builder, sensorName): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(sensorName), 0)
def AddSensorName(builder, sensorName):
    return SensorAddSensorName(builder, sensorName)
def SensorAddSensorTypes(builder, sensorTypes): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(sensorTypes), 0)
def AddSensorTypes(builder, sensorTypes):
    return SensorAddSensorTypes(builder, sensorTypes)
def SensorAddCameraType(builder, cameraType): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(cameraType), 0)
def AddCameraType(builder, cameraType):
    return SensorAddCameraType(builder, cameraType)
def SensorAddBoundingBox(builder, boundingBox): builder.PrependBoolSlot(3, boundingBox, 0)
def AddBoundingBox(builder, boundingBox):
    return SensorAddBoundingBox(builder, boundingBox)
def SensorAddTrackActorTypes(builder, trackActorTypes): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(trackActorTypes), 0)
def AddTrackActorTypes(builder, trackActorTypes):
    return SensorAddTrackActorTypes(builder, trackActorTypes)
def SensorStartTrackActorTypesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartTrackActorTypesVector(builder, numElems):
    return SensorStartTrackActorTypesVector(builder, numElems)
def SensorAddShowBoundingBoxes(builder, showBoundingBoxes): builder.PrependBoolSlot(5, showBoundingBoxes, 0)
def AddShowBoundingBoxes(builder, showBoundingBoxes):
    return SensorAddShowBoundingBoxes(builder, showBoundingBoxes)
def SensorAddWorldSensor(builder, worldSensor): builder.PrependBoolSlot(6, worldSensor, 0)
def AddWorldSensor(builder, worldSensor):
    return SensorAddWorldSensor(builder, worldSensor)
def SensorAddSensorLocation(builder, sensorLocation): builder.PrependStructSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(sensorLocation), 0)
def AddSensorLocation(builder, sensorLocation):
    return SensorAddSensorLocation(builder, sensorLocation)
def SensorAddSensorRotation(builder, sensorRotation): builder.PrependStructSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(sensorRotation), 0)
def AddSensorRotation(builder, sensorRotation):
    return SensorAddSensorRotation(builder, sensorRotation)
def SensorAddSensorResolution(builder, sensorResolution): builder.PrependStructSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(sensorResolution), 0)
def AddSensorResolution(builder, sensorResolution):
    return SensorAddSensorResolution(builder, sensorResolution)
def SensorAddFrameOfReference(builder, frameOfReference): builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(frameOfReference), 0)
def AddFrameOfReference(builder, frameOfReference):
    return SensorAddFrameOfReference(builder, frameOfReference)
def SensorAddFov(builder, fov): builder.PrependFloat64Slot(11, fov, 0.0)
def AddFov(builder, fov):
    return SensorAddFov(builder, fov)
def SensorAddAttachToActor(builder, attachToActor): builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(attachToActor), 0)
def AddAttachToActor(builder, attachToActor):
    return SensorAddAttachToActor(builder, attachToActor)
def SensorAddRadius(builder, radius): builder.PrependFloat64Slot(13, radius, 0.0)
def AddRadius(builder, radius):
    return SensorAddRadius(builder, radius)
def SensorAddOcclusion(builder, occlusion): builder.PrependBoolSlot(14, occlusion, 0)
def AddOcclusion(builder, occlusion):
    return SensorAddOcclusion(builder, occlusion)
def SensorEnd(builder): return builder.EndObject()
def End(builder):
    return SensorEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Location(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 24

    # Location
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Location
    def X(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Location
    def Y(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # Location
    def Z(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16))

def CreateLocation(builder, x, y, z):
    builder.Prep(8, 24)
    builder.PrependFloat64(z)
    builder.PrependFloat64(y)
    builder.PrependFloat64(x)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Resolution(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 8

    # Resolution
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Resolution
    def Width(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Resolution
    def Height(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))

def CreateResolution(builder, width, height):
    builder.Prep(4, 8)
    builder.PrependInt32(height)
    builder.PrependInt32(width)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ServerHandshake(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ServerHandshake()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsServerHandshake(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ServerHandshakeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # ServerHandshake
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ServerHandshake
    def SystemName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ServerHandshake
    def ModelName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def ServerHandshakeStart(builder): builder.StartObject(2)
def Start(builder):
    return ServerHandshakeStart(builder)
def ServerHandshakeAddSystemName(builder, systemName): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(systemName), 0)
def AddSystemName(builder, systemName):
    return ServerHandshakeAddSystemName(builder, systemName)
def ServerHandshakeAddModelName(builder, modelName): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(modelName), 0)
def AddModelName(builder, modelName):
    return ServerHandshakeAddModelName(builder, modelName)
def ServerHandshakeEnd(builder): return builder.EndObject()
def End(builder):
    return ServerHandshakeEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class WorldParameters(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = WorldParameters()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsWorldParameters(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def WorldParametersBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # WorldParameters
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # WorldParameters
    def Carlatown(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def WorldParametersStart(builder): builder.StartObject(1)
def Start(builder):
    return WorldParametersStart(builder)
def WorldParametersAddCarlatown(builder, carlatown): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(carlatown), 0)
def AddCarlatown(builder, carlatown):
    return WorldParametersAddCarlatown(builder, carlatown)
def WorldParametersEnd(builder): return builder.EndObject()
def End(builder):
    return WorldParametersEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class BoundingBox(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BoundingBox()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsBoundingBox(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def BoundingBoxBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # BoundingBox
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # BoundingBox
    def ActorType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # BoundingBox
    def Ids(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # BoundingBox
    def IdsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # BoundingBox
    def IdsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # BoundingBox
    def IdsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # BoundingBox
    def Coordinates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # BoundingBox
    def CoordinatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # BoundingBox
    def CoordinatesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # BoundingBox
    def State(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # BoundingBox
    def StateLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # BoundingBox
    def StateIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # BoundingBox
    def Occlusion(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # BoundingBox
    def OcclusionAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # BoundingBox
    def OcclusionLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # BoundingBox
    def OcclusionIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

def BoundingBoxStart(builder): builder.StartObject(5)
def Start(builder):
    return BoundingBoxStart(builder)
def BoundingBoxAddActorType(builder, actorType): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(actorType), 0)
def AddActorType(builder, actorType):
    return BoundingBoxAddActorType(builder, actorType)
def BoundingBoxAddIds(builder, ids): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(ids), 0)
def AddIds(builder, ids):
    return BoundingBoxAddIds(builder, ids)
def BoundingBoxStartIdsVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def StartIdsVector(builder, numElems):
    return BoundingBoxStartIdsVector(builder, numElems)
def BoundingBoxAddCoordinates(builder, coordinates): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(coordinates), 0)
def AddCoordinates(builder, coordinates):
    return BoundingBoxAddCoordinates(builder, coordinates)
def BoundingBoxStartCoordinatesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartCoordinatesVector(builder, numElems):
    return BoundingBoxStartCoordinatesVector(builder, numElems)
def BoundingBoxAddState(builder, state): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(state), 0)
def AddState(builder, state):
    return BoundingBoxAddState(builder, state)
def BoundingBoxStartStateVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartStateVector(builder, numElems):
    return BoundingBoxStartStateVector(builder, numElems)
def BoundingBoxAddOcclusion(builder, occlusion): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(occlusion), 0)
def AddOcclusion(builder, occlusion):
    return BoundingBoxAddOcclusion(builder, occlusion)
def BoundingBoxStartOcclusionVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def StartOcclusionVector(builder, numElems):
    return BoundingBoxStartOcclusionVector(builder, numElems)
def BoundingBoxEnd(builder): return builder.EndObject()
def End(builder):
    return BoundingBoxEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SensorsDataDict(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SensorsDataDict()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSensorsDataDict(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def SensorsDataDictBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # SensorsDataDict
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SensorsDataDict
    def SensorsData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from communications.isp.SensorData import SensorData
            obj = SensorData()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SensorsDataDict
    def SensorsDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SensorsDataDict
    def SensorsDataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

def SensorsDataDictStart(builder): builder.StartObject(1)
def Start(builder):
    return SensorsDataDictStart(builder)
def SensorsDataDictAddSensorsData(builder, sensorsData): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(sensorsData), 0)
def AddSensorsData(builder, sensorsData):
    return SensorsDataDictAddSensorsData(builder, sensorsData)
def SensorsDataDictStartSensorsDataVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartSensorsDataVector(builder, numElems):
    return SensorsDataDictStartSensorsDataVector(builder, numElems)
def SensorsDataDictEnd(builder): return builder.EndObject()
def End(builder):
    return SensorsDataDictEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ForwardGears(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ForwardGears()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsForwardGears(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ForwardGearsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # ForwardGears
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ForwardGears
    def Ratio(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # ForwardGears
    def DownRatio(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # ForwardGears
    def UpRatio(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

def ForwardGearsStart(builder): builder.StartObject(3)
def Start(builder):
    return ForwardGearsStart(builder)
def ForwardGearsAddRatio(builder, ratio): builder.PrependFloat64Slot(0, ratio, 0.0)
def AddRatio(builder, ratio):
    return ForwardGearsAddRatio(builder, ratio)
def ForwardGearsAddDownRatio(builder, downRatio): builder.PrependFloat64Slot(1, downRatio, 0.0)
def AddDownRatio(builder, downRatio):
    return ForwardGearsAddDownRatio(builder, downRatio)
def ForwardGearsAddUpRatio(builder, upRatio): builder.PrependFloat64Slot(2, upRatio, 0.0)
def AddUpRatio(builder, upRatio):
    return ForwardGearsAddUpRatio(builder, upRatio)
def ForwardGearsEnd(builder): return builder.EndObject()
def End(builder):
    return ForwardGearsEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Info(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Info()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsInfo(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def InfoBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # Info
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Info
    def Invasion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Info
    def Collision(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Info
    def Gear(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Info
    def ExpertAction(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Action import Action
            obj = Action()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Info
    def Outcome(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Info
    def NextWaypointDistance(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # Info
    def DistanceToGoal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # Info
    def WaypointsFound(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Info
    def WaypointsTotal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Info
    def WaypointsRemaining(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def InfoStart(builder): builder.StartObject(10)
def Start(builder):
    return InfoStart(builder)
def InfoAddInvasion(builder, invasion): builder.PrependBoolSlot(0, invasion, 0)
def AddInvasion(builder, invasion):
    return InfoAddInvasion(builder, invasion)
def InfoAddCollision(builder, collision): builder.PrependBoolSlot(1, collision, 0)
def AddCollision(builder, collision):
    return InfoAddCollision(builder, collision)
def InfoAddGear(builder, gear): builder.PrependInt32Slot(2, gear, 0)
def AddGear(builder, gear):
    return InfoAddGear(builder, gear)
def InfoAddExpertAction(builder, expertAction): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(expertAction), 0)
def AddExpertAction(builder, expertAction):
    return InfoAddExpertAction(builder, expertAction)
def InfoAddOutcome(builder, outcome): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(outcome), 0)
def AddOutcome(builder, outcome):
    return InfoAddOutcome(builder, outcome)
def InfoAddNextWaypointDistance(builder, nextWaypointDistance): builder.PrependFloat64Slot(5, nextWaypointDistance, 0.0)
def AddNextWaypointDistance(builder, nextWaypointDistance):
    return InfoAddNextWaypointDistance(builder, nextWaypointDistance)
def InfoAddDistanceToGoal(builder, distanceToGoal): builder.PrependFloat64Slot(6, distanceToGoal, 0.0)
def AddDistanceToGoal(builder, distanceToGoal):
    return InfoAddDistanceToGoal(builder, distanceToGoal)
def InfoAddWaypointsFound(builder, waypointsFound): builder.PrependInt32Slot(7, waypointsFound, 0)
def AddWaypointsFound(builder, waypointsFound):
    return InfoAddWaypointsFound(builder, waypointsFound)
def InfoAddWaypointsTotal(builder, waypointsTotal): builder.PrependInt32Slot(8, waypointsTotal, 0)
def AddWaypointsTotal(builder, waypointsTotal):
    return InfoAddWaypointsTotal(builder, waypointsTotal)
def InfoAddWaypointsRemaining(builder, waypointsRemaining): builder.PrependInt32Slot(9, waypointsRemaining, 0)
def AddWaypointsRemaining(builder, waypointsRemaining):
    return InfoAddWaypointsRemaining(builder, waypointsRemaining)
def InfoEnd(builder): return builder.EndObject()
def End(builder):
    return InfoEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Initialize(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Initialize()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsInitialize(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def InitializeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # Initialize
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Initialize
    def Scenario(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Initialize
    def ScenarioParameters(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.ScenarioParameters import ScenarioParameters
            obj = ScenarioParameters()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Initialize
    def WorldParameters(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.WorldParameters import WorldParameters
            obj = WorldParameters()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Initialize
    def VehiclePhysics(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.VehiclePhysics import VehiclePhysics
            obj = VehiclePhysics()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Initialize
    def Sensors(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.SensorsDict import SensorsDict
            obj = SensorsDict()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def InitializeStart(builder): builder.StartObject(5)
def Start(builder):
    return InitializeStart(builder)
def InitializeAddScenario(builder, scenario): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(scenario), 0)
def AddScenario(builder, scenario):
    return InitializeAddScenario(builder, scenario)
def InitializeAddScenarioParameters(builder, scenarioParameters): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(scenarioParameters), 0)
def AddScenarioParameters(builder, scenarioParameters):
    return InitializeAddScenarioParameters(builder, scenarioParameters)
def InitializeAddWorldParameters(builder, worldParameters): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(worldParameters), 0)
def AddWorldParameters(builder, worldParameters):
    return InitializeAddWorldParameters(builder, worldParameters)
def InitializeAddVehiclePhysics(builder, vehiclePhysics): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(vehiclePhysics), 0)
def AddVehiclePhysics(builder, vehiclePhysics):
    return InitializeAddVehiclePhysics(builder, vehiclePhysics)
def InitializeAddSensors(builder, sensors): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(sensors), 0)
def AddSensors(builder, sensors):
    return InitializeAddSensors(builder, sensors)
def InitializeEnd(builder): return builder.EndObject()
def End(builder):
    return InitializeEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class WheelPhysics(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = WheelPhysics()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsWheelPhysics(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def WheelPhysicsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # WheelPhysics
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # WheelPhysics
    def TireFriction(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 2.0

    # WheelPhysics
    def DampingRate(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 1.5

    # WheelPhysics
    def MaxSteerAngle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 70.0

    # WheelPhysics
    def Radius(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # WheelPhysics
    def MaxBrakeTorque(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # WheelPhysics
    def MaxHandbrakeTorque(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # WheelPhysics
    def LongStiffValue(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 1000.0

    # WheelPhysics
    def LatStiffMaxLoad(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # WheelPhysics
    def LatStiffValue(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

def WheelPhysicsStart(builder): builder.StartObject(9)
def Start(builder):
    return WheelPhysicsStart(builder)
def WheelPhysicsAddTireFriction(builder, tireFriction): builder.PrependFloat64Slot(0, tireFriction, 2.0)
def AddTireFriction(builder, tireFriction):
    return WheelPhysicsAddTireFriction(builder, tireFriction)
def WheelPhysicsAddDampingRate(builder, dampingRate): builder.PrependFloat64Slot(1, dampingRate, 1.5)
def AddDampingRate(builder, dampingRate):
    return WheelPhysicsAddDampingRate(builder, dampingRate)
def WheelPhysicsAddMaxSteerAngle(builder, maxSteerAngle): builder.PrependFloat64Slot(2, maxSteerAngle, 70.0)
def AddMaxSteerAngle(builder, maxSteerAngle):
    return WheelPhysicsAddMaxSteerAngle(builder, maxSteerAngle)
def WheelPhysicsAddRadius(builder, radius): builder.PrependFloat64Slot(3, radius, 0.0)
def AddRadius(builder, radius):
    return WheelPhysicsAddRadius(builder, radius)
def WheelPhysicsAddMaxBrakeTorque(builder, maxBrakeTorque): builder.PrependFloat64Slot(4, maxBrakeTorque, 0.0)
def AddMaxBrakeTorque(builder, maxBrakeTorque):
    return WheelPhysicsAddMaxBrakeTorque(builder, maxBrakeTorque)
def WheelPhysicsAddMaxHandbrakeTorque(builder, maxHandbrakeTorque): builder.PrependFloat64Slot(5, maxHandbrakeTorque, 0.0)
def AddMaxHandbrakeTorque(builder, maxHandbrakeTorque):
    return WheelPhysicsAddMaxHandbrakeTorque(builder, maxHandbrakeTorque)
def WheelPhysicsAddLongStiffValue(builder, longStiffValue): builder.PrependFloat64Slot(6, longStiffValue, 1000.0)
def AddLongStiffValue(builder, longStiffValue):
    return WheelPhysicsAddLongStiffValue(builder, longStiffValue)
def WheelPhysicsAddLatStiffMaxLoad(builder, latStiffMaxLoad): builder.PrependFloat64Slot(7, latStiffMaxLoad, 0.0)
def AddLatStiffMaxLoad(builder, latStiffMaxLoad):
    return WheelPhysicsAddLatStiffMaxLoad(builder, latStiffMaxLoad)
def WheelPhysicsAddLatStiffValue(builder, latStiffValue): builder.PrependFloat64Slot(8, latStiffValue, 0.0)
def AddLatStiffValue(builder, latStiffValue):
    return WheelPhysicsAddLatStiffValue(builder, latStiffValue)
def WheelPhysicsEnd(builder): return builder.EndObject()
def End(builder):
    return WheelPhysicsEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Action(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 16

    # Action
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Action
    def Acceleration(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Action
    def Steering(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateAction(builder, acceleration, steering):
    builder.Prep(8, 16)
    builder.PrependFloat64(steering)
    builder.PrependFloat64(acceleration)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SensorData(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SensorData()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSensorData(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def SensorDataBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # SensorData
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SensorData
    def SensorName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # SensorData
    def Image(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SensorData
    def BoundingBox(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from communications.isp.BoundingBox import BoundingBox
            obj = BoundingBox()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SensorData
    def BoundingBoxLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SensorData
    def BoundingBoxIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def SensorDataStart(builder): builder.StartObject(3)
def Start(builder):
    return SensorDataStart(builder)
def SensorDataAddSensorName(builder, sensorName): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(sensorName), 0)
def AddSensorName(builder, sensorName):
    return SensorDataAddSensorName(builder, sensorName)
def SensorDataAddImage(builder, image): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(image), 0)
def AddImage(builder, image):
    return SensorDataAddImage(builder, image)
def SensorDataAddBoundingBox(builder, boundingBox): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(boundingBox), 0)
def AddBoundingBox(builder, boundingBox):
    return SensorDataAddBoundingBox(builder, boundingBox)
def SensorDataStartBoundingBoxVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StartBoundingBoxVector(builder, numElems):
    return SensorDataStartBoundingBoxVector(builder, numElems)
def SensorDataEnd(builder): return builder.EndObject()
def End(builder):
    return SensorDataEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Command(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Command()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCommand(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def CommandBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # Command
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Command
    def Cmd(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Command
    def Step(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Step import Step
            obj = Step()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Command
    def TensorArgs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def CommandStart(builder): builder.StartObject(3)
def Start(builder):
    return CommandStart(builder)
def CommandAddCmd(builder, cmd): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(cmd), 0)
def AddCmd(builder, cmd):
    return CommandAddCmd(builder, cmd)
def CommandAddStep(builder, step): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(step), 0)
def AddStep(builder, step):
    return CommandAddStep(builder, step)
def CommandAddTensorArgs(builder, tensorArgs): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(tensorArgs), 0)
def AddTensorArgs(builder, tensorArgs):
    return CommandAddTensorArgs(builder, tensorArgs)
def CommandEnd(builder): return builder.EndObject()
def End(builder):
    return CommandEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ScenarioParameters(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ScenarioParameters()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsScenarioParameters(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ScenarioParametersBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # ScenarioParameters
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ScenarioParameters
    def EgoStart(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ScenarioParameters
    def MaxNumAgents(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScenarioParameters
    def MaxScenLen(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScenarioParameters
    def CameraLocationVariation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Location import Location
            obj = Location()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ScenarioParameters
    def CameraOrientationVariation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from communications.isp.Rotation import Rotation
            obj = Rotation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def ScenarioParametersStart(builder): builder.StartObject(5)
def Start(builder):
    return ScenarioParametersStart(builder)
def ScenarioParametersAddEgoStart(builder, egoStart): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(egoStart), 0)
def AddEgoStart(builder, egoStart):
    return ScenarioParametersAddEgoStart(builder, egoStart)
def ScenarioParametersAddMaxNumAgents(builder, maxNumAgents): builder.PrependInt32Slot(1, maxNumAgents, 0)
def AddMaxNumAgents(builder, maxNumAgents):
    return ScenarioParametersAddMaxNumAgents(builder, maxNumAgents)
def ScenarioParametersAddMaxScenLen(builder, maxScenLen): builder.PrependInt32Slot(2, maxScenLen, 0)
def AddMaxScenLen(builder, maxScenLen):
    return ScenarioParametersAddMaxScenLen(builder, maxScenLen)
def ScenarioParametersAddCameraLocationVariation(builder, cameraLocationVariation): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(cameraLocationVariation), 0)
def AddCameraLocationVariation(builder, cameraLocationVariation):
    return ScenarioParametersAddCameraLocationVariation(builder, cameraLocationVariation)
def ScenarioParametersAddCameraOrientationVariation(builder, cameraOrientationVariation): builder.PrependStructSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(cameraOrientationVariation), 0)
def AddCameraOrientationVariation(builder, cameraOrientationVariation):
    return ScenarioParametersAddCameraOrientationVariation(builder, cameraOrientationVariation)
def ScenarioParametersEnd(builder): return builder.EndObject()
def End(builder):
    return ScenarioParametersEnd(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: isp

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class State(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = State()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsState(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def StateBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x49\x53\x50\x46", size_prefixed=size_prefixed)

    # State
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # State
    def Obs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Obs import Obs
            obj = Obs()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # State
    def Reward(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # State
    def Done(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # State
    def Info(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from communications.isp.Info import Info
            obj = Info()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def StateStart(builder): builder.StartObject(4)
def Start(builder):
    return StateStart(builder)
def StateAddObs(builder, obs): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(obs), 0)
def AddObs(builder, obs):
    return StateAddObs(builder, obs)
def StateAddReward(builder, reward): builder.PrependFloat64Slot(1, reward, 0.0)
def AddReward(builder, reward):
    return StateAddReward(builder, reward)
def StateAddDone(builder, done): builder.PrependBoolSlot(2, done, 0)
def AddDone(builder, done):
    return StateAddDone(builder, done)
def StateAddInfo(builder, info): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(info), 0)
def AddInfo(builder, info):
    return StateAddInfo(builder, info)
def StateEnd(builder): return builder.EndObject()
def End(builder):
    return StateEnd(builder)